from unittest.mock import Mock, patch

import pytest
from aviary.env import DummyEnv
from aviary.message import Message
from aviary.tools import Tool, ToolRequestMessage
from pytest_subtests import SubTests

from ldp.agent import ReActAgent
from ldp.graph import OpResult
from ldp.graph.modules import (
    ReActModule,
    ReflectModule,
    ReflectModuleConfig,
    parse_message,
)

from . import CILLMModelNames


@pytest.mark.asyncio
@pytest.mark.vcr
async def test_reflect_module() -> None:
    config = ReflectModuleConfig(
        llm_model={"model": CILLMModelNames.ANTHROPIC.value, "temperature": 0}
    )  # Lower temperature for more deterministic responses
    reflect_module = ReflectModule(config)
    context = "I am happy. How do I feel?"
    response = "You are sad."
    result = await reflect_module(context, response)
    assert isinstance(result, OpResult)
    assert len(result.value) > 0
    assert result.value != response
    # Check both emotions to work around LLM not responding with "happy"
    # For example: "It sounds like you are feeling joyful."
    assert "happy" in result.value or "sad" not in result.value


@pytest.fixture(name="mock_tools")
def fixture_mock_tools() -> list[Tool]:
    return [Mock(spec=Tool, name=f"Tool{i}") for i in range(3)]


class TestReActModule:
    @pytest.mark.asyncio
    async def test_templating(self, dummy_env: DummyEnv) -> None:
        obs, tools = await dummy_env.reset()
        module = ReActModule(ReActAgent.model_fields["llm_model"].default)
        with patch(
            "ldp.graph.common_ops.LLMCallOp.forward",
            return_value=ToolRequestMessage(
                role="assistant",
                content=f"Action: {tools[0].info.name}\nAction Input: stub",
            ),
        ) as mock_forward:
            await module(obs, tools=tools)
        mock_forward.assert_awaited_once()
        assert mock_forward.await_args
        assert mock_forward.await_args[1]["msgs"][0] == Message(
            role="system",
            content=(
                "Answer the following questions as best you can. You have access to the"
                " following tools:\n\nNAME: print_story\n\nSYNOPSIS:\n   "
                " print_story(string story)\n\nDESCRIPTION:\n    Print a"
                " story.\n\nPARAMETERS:\n    story (string): Story to print.\n\nNAME:"
                " cast_float\n\nSYNOPSIS:\n    cast_float(string"
                " x)\n\nDESCRIPTION:\n    Cast the input argument x to a"
                " float.\n\nPARAMETERS:\n    x (string): No description"
                " provided.\n\nNAME: cast_int\n\nSYNOPSIS:\n   "
                " cast_int(number x)\n\nDESCRIPTION:\n    Cast the input argument x to"
                " an integer.\n\nPARAMETERS:\n    x (number): No description"
                " provided.\n\nUse the following format:\n\nThought: you should"
                " always think about what to do\nAction: the action to take, should be"
                " one of [print_story, cast_float, cast_int]\nAction Input: comma"
                " separated list of inputs to action as python tuple\nObservation: the"
                " result of the action\n... (this Thought/Action/Action"
                " Input/Observation can repeat N times)\n\nExample:\n\nThought: I need"
                ' to use the get_weather tool\nAction: get_weather\nAction Input: "New'
                ' York", 7\nObservation: The 7 day forecast for New York is [...]'
            ),
        )

    @pytest.mark.asyncio
    async def test_react_parse(self, subtests: SubTests) -> None:  # noqa: PLR0915, C901
        def zero_arg() -> int:
            """A test function."""
            return 0

        def one_arg(x: int) -> int:
            """A test function.

            Args:
                x: x
            """
            return x

        def one_string_arg(x: str) -> str:
            """A test function.

            Args:
                x: x
            """
            return x

        def one_alphanumeric_string_arg(x: str) -> str:
            """A test function.

            Args:
                x: x
            """
            return x

        def two_arg(x: int, y: str) -> int:
            """A test function.

            Args:
                x: x
                y: y
            """
            return x + len(y)

        def one_optional_arg(x: int = 1) -> int:
            """A test function.

            Args:
                x: x with a default.
            """
            return x

        def mixed_required_args(x: int, y: int = 1) -> int:
            """A test function.

            Args:
                x: x
                y: y
            """
            return x + y

        def two_string_args(x: str, y: str) -> str:
            """A test function.

            Args:
                x: x
                y: y
            """
            return x + y

        def three_string_args(x: str, y: str, z: str) -> str:
            """A test function.

            Args:
                x: x
                y: y
                z: z
            """
            return x + y + z

        def mixed_string_int_args(x: str, y: int, z: str) -> str:
            """A test function.

            Args:
                x: x
                y: y
                z: z
            """
            return (x + z) * y

        def complex_optionals(
            x: str, y: str, a: int | None = None, b: int | None = None
        ) -> str:
            """A complex test function with optionals.

            Args:
                x: x
                y: y
                a: a
                b: b
            """
            return x + y + str(a) + str(b)

        tools = [
            Tool.from_function(zero_arg),
            Tool.from_function(one_arg),
            Tool.from_function(one_string_arg),
            Tool.from_function(one_alphanumeric_string_arg),
            Tool.from_function(two_arg),
            Tool.from_function(one_optional_arg),
            Tool.from_function(mixed_required_args),
            Tool.from_function(two_string_args),
            Tool.from_function(three_string_args),
            Tool.from_function(mixed_string_int_args),
            Tool.from_function(complex_optionals),
        ]

        with subtests.test("no args"):
            message = Message(
                content="""

        Of course, I would be happy to follow your format.

        This will be fun!

        I cannot wait to begin!!!

        I had a thought? What if I don't follow it

        OK... I will:

        Thought: you should always think about what to do
        Action: zero_arg
        Action Input:
        Observation: the result of the action

        """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "zero_arg"
            assert not tool_call.function.arguments

        with subtests.test("one arg"):
            message = Message(
                content="""

        Of course, I would be happy to follow your format.

            Thought:
            Action: one_arg
            Action Input: 1
        """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "one_arg"
            assert tool_call.function.arguments == {"x": 1}

        with subtests.test("one string arg"):
            message = Message(
                content="""

        Of course, I would be happy to follow your format.

            Thought:
            Action: one_string_arg
            Action Input: Bertrand Russell
        """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "one_string_arg"
            assert tool_call.function.arguments == {"x": "Bertrand Russell"}

        with subtests.test("one alphanumeric string arg"):
            message = Message(
                content="""

        Of course, I would be happy to follow your format.

            Thought:
            Action: one_alphanumeric_string_arg
            Action Input: R2D2 from Star Wars
        """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "one_alphanumeric_string_arg"
            assert tool_call.function.arguments == {"x": "R2D2 from Star Wars"}

        with subtests.test("two args"):
            message = Message(
                content="""
            Hey, here's some code:

            ```js
            console.log("hello world!"!)
            ```

            just thought you'd want to know!!!!

            Thought: A long thought
            what do I do next???
            Oh right
            Note: spacing below is part of test
            Action:   two_arg
            Action Input: 1, 2
            """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "two_arg"
            assert tool_call.function.arguments == {"x": 1, "y": 2}

        with subtests.test("nested args"):
            # check that parser can handle if the action input
            # is itself already a tuple
            message = Message(
                content="""
            Hey, here's some code:

            ```js
            console.log("hello world!"!)
            ```

            just thought you'd want to know!!!!

            Thought: A long thought
            what do I do next???
            Oh right
            Action:one_arg
            Action Input: (1,)
            """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "one_arg"
            assert tool_call.function.arguments == {"x": 1}

        with subtests.test("optional arg"):
            message = Message(
                content="""
            Thought: A long thought
            what do I do next???
            Oh right
            Action: one_optional_arg
            Action Input: ()
            """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "one_optional_arg"
            assert not tool_call.function.arguments

        with subtests.test("mixed args"):
            message = Message(
                content="""
            Thought: A long thought
            what do I do next???
            Oh right
            Action: mixed_required_args
            Action Input: (1,)
            """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "mixed_required_args"
            assert tool_call.function.arguments == {"x": 1}

        with subtests.test("two mixed args"):
            message = Message(
                content="""
            Thought: A long thought
            what do I do next???
            Oh right
            Action: mixed_required_args
            Action Input: (1,2)
            """
            )
            tool_call = parse_message(message, tools).tool_calls[0]
            assert tool_call.function.name == "mixed_required_args"
            assert tool_call.function.arguments == {"x": 1, "y": 2}

        with subtests.test("two string args"):
            message = Message(
                content="""
            Thought: A long thought
            what do I do next???
            Oh right
            Action: two_string_args
            Action Input: (Bertrand Russell, Friedrich Nietzsche)
            """
            )
            tool_request_message = parse_message(message, tools)
            assert isinstance(tool_request_message, ToolRequestMessage)
            assert tool_request_message.tool_calls[0].function.name == "two_string_args"
            assert tool_request_message.tool_calls[0].function.arguments == {
                "x": "Bertrand Russell",
                "y": "Friedrich Nietzsche",
            }

        with subtests.test("three string args"):
            message = Message(
                content="""
            Thought: A long thought
            what do I do next???
            Oh right
            Action: three_string_args
            Action Input: (Bertrand Russell, Friedrich Nietzsche, Immanuel Kant)
            """
            )
            tool_request_message = parse_message(message, tools)
            assert isinstance(tool_request_message, ToolRequestMessage)
            assert (
                tool_request_message.tool_calls[0].function.name == "three_string_args"
            )
            assert tool_request_message.tool_calls[0].function.arguments == {
                "x": "Bertrand Russell",
                "y": "Friedrich Nietzsche",
                "z": "Immanuel Kant",
            }

        with subtests.test("mixed string and integer args"):
            message = Message(
                content="""
            Thought: A long thought
            what do I do next???
            Oh right
            Action: mixed_string_int_args
            Action Input: (Bertrand Russell, 2, Immanuel Kant)
            """
            )
            tool_request_message = parse_message(message, tools)
            assert isinstance(tool_request_message, ToolRequestMessage)
            assert (
                tool_request_message.tool_calls[0].function.name
                == "mixed_string_int_args"
            )
            assert tool_request_message.tool_calls[0].function.arguments == {
                "x": "Bertrand Russell",
                "y": 2,
                "z": "Immanuel Kant",
            }

        with subtests.test("Attempting to put it into python ticks"):
            text = """
                ```python
                Thought: seq-2f794d237c85 contains the DsRed gene, a variant of RFP. I need to extract this gene and clone it into an E. coli-compatible vector. First, I'll slice out the RFP gene.

                Action: complex_optionals
                Action Input: ann-seq-5d7de2e1-f32a, EGE3L, 1697, 2372
                ```"""

            message = Message(content=text)
            tool_request_message = parse_message(message, tools)

        with subtests.test("A bunch of actions still only calls one tool"):
            text = """
                ```python
                Thought: seq-2f794d237c85 contains the DsRed gene, a variant of RFP. I need to extract this gene and clone it into an E. coli-compatible vector. First, I'll slice out the RFP gene.

                Action: complex_optionals
                Action Input: ann-seq-5d7de2e1-f32a, EGE3L, 1697, 2372

                Action: complex_optionals
                Action Input: ann-seq-5d7de2e1-f32a, EGE3L, 1697, 2372

                Action: complex_optionals
                Action Input: ann-seq-5d7de2e1-f32a, EGE3L, 1697, 2372
                ```"""

            message = Message(content=text)
            tool_request_message = parse_message(message, tools)
            assert len(tool_request_message.tool_calls) == 1

        with subtests.test("a single integer arg"):
            _, tools = await DummyEnv().reset()
            test_message = Message(
                role="assistant",
                content="Thought: I think.\nAction: cast_int\nAction Input: (1.1)",
            )
            parsed_message = parse_message(test_message, tools)
            assert isinstance(parsed_message, ToolRequestMessage)
            assert parsed_message.tool_calls[0].function.name == "cast_int"
            assert parsed_message.tool_calls[0].function.arguments == {"x": 1.1}



from typing import ClassVar
from unittest.mock import Mock

import litellm
import numpy as np
import pytest
from aviary.env import DummyEnv
from aviary.message import Message
from aviary.tools import Tool, ToolRequestMessage
from pydantic import BaseModel, Field

from ldp.llms import (
    JSONSchemaValidationError,
    LLMModel,
    LLMResult,
    MultipleCompletionLLMModel,
    validate_json_completion,
)

from . import CILLMModelNames


def test_json_schema_validation() -> None:
    # Invalid JSON
    mock_completion1 = Mock()
    mock_completion1.choices = [Mock()]
    mock_completion1.choices[0].message.content = "not a json"
    # Invalid schema
    mock_completion2 = Mock()
    mock_completion2.choices = [Mock()]
    mock_completion2.choices[0].message.content = '{"name": "John", "age": "nan"}'
    # Valid schema
    mock_completion3 = Mock()
    mock_completion3.choices = [Mock()]
    mock_completion3.choices[0].message.content = '{"name": "John", "age": 30}'

    class DummyModel(BaseModel):
        name: str
        age: int

    with pytest.raises(JSONSchemaValidationError):
        validate_json_completion(mock_completion1, DummyModel)
    with pytest.raises(JSONSchemaValidationError):
        validate_json_completion(mock_completion2, DummyModel)
    validate_json_completion(mock_completion3, DummyModel)


@pytest.mark.parametrize(
    "model_name", ["gpt-3.5-turbo", CILLMModelNames.ANTHROPIC.value]
)
@pytest.mark.asyncio
async def test_achat(model_name: str) -> None:
    model = LLMModel(name=model_name)
    response = await model.achat(
        messages=[
            Message(content="What are three things I should do today?"),
        ]
    )

    assert len(response.choices) == 1

    # Check we can iterate through the response
    async for chunk in await model.achat_iter(
        messages=[
            Message(content="What are three things I should do today?"),
        ]
    ):
        assert len(chunk.choices) == 1


@pytest.mark.parametrize(
    "model_name", [CILLMModelNames.OPENAI.value, CILLMModelNames.ANTHROPIC.value]
)
@pytest.mark.flaky(reruns=3, only_on=[litellm.exceptions.APIConnectionError])
@pytest.mark.asyncio
async def test_tools(dummy_env: DummyEnv, model_name: str) -> None:
    model = LLMModel(name=model_name)
    messages = [
        Message(content="What are three things I should do today?"),
    ]

    def get_todo_list(n: int) -> str:
        """Get todo list for today.

        Args:
            n: number of items to return
        """
        return "\n".join(["Go for a walk", "Read a book", "Call a friend"][:n])

    tool = Tool.from_function(get_todo_list)
    dummy_env.tools = [tool]
    result = await model.call(messages, tools=dummy_env.tools)
    assert result.completion_count > 0

    # try specifying tool choice
    result = await model.call(messages, tools=dummy_env.tools, tool_choice=tool)
    assert result.completion_count > 0, "Tool choice failed to execute tool"
    assert result.messages
    (tool_request_message,) = result.messages
    assert isinstance(tool_request_message, ToolRequestMessage)

    new_messages = await dummy_env.exec_tool_calls(tool_request_message)
    (new_message,) = new_messages
    assert new_message.content == "Go for a walk\nRead a book\nCall a friend"
    assert new_message.tool_call_id == tool_request_message.tool_calls[0].id

    def get_todo_list_no_args() -> str:
        """Get todo list for today."""
        return "\n".join(["Go for a walk", "Read a book", "Call a friend"])

    tool = Tool.from_function(get_todo_list_no_args)
    dummy_env.tools = [tool]
    result = await model.call(messages, tools=dummy_env.tools)
    assert result.completion_count > 0
    assert result.messages
    (tool_request_message,) = result.messages
    assert isinstance(tool_request_message, ToolRequestMessage)

    new_messages = await dummy_env.exec_tool_calls(tool_request_message)
    (new_message,) = new_messages
    assert new_message.content == "Go for a walk\nRead a book\nCall a friend"
    assert new_message.tool_call_id == tool_request_message.tool_calls[0].id

    # ok now try with multiple functions
    messages = [
        Message(
            content=(
                "What items will I have time to accomplish on my todo list today based"
                " on my calendar?"
            )
        ),
    ]

    def get_calendar() -> str:
        """Get text version of calendar for today."""
        return "9:00am Wake-up\n10:00pm Go to bed\n"

    tool2 = Tool.from_function(get_calendar)
    dummy_env.tools = [tool, tool2]
    result = await model.call(messages, tools=dummy_env.tools)
    assert result.messages
    (tool_request_message,) = result.messages
    assert isinstance(tool_request_message, ToolRequestMessage)
    new_messages = await dummy_env.exec_tool_calls(tool_request_message)
    if model_name.startswith("claude"):
        # Anthropic not always so smart
        assert 1 <= len(new_messages) <= 2
    else:
        assert len(new_messages) == 2

    # ok now try continuation - I AM NOT SURE IF THIS IS VALID?
    # TODO: - supported on openai, but not litellm
    # messages = messages + result.messages + new_messages
    # result = await model.call(messages)


class DummyOutputSchema(BaseModel):
    name: str
    age: int


class TestMultipleCompletionLLMModel:
    NUM_COMPLETIONS: ClassVar[int] = 2
    DEFAULT_CONFIG: ClassVar[dict] = {"n": NUM_COMPLETIONS}
    MODEL_CLS: ClassVar[type[MultipleCompletionLLMModel]] = MultipleCompletionLLMModel

    async def call_model(
        self, model: MultipleCompletionLLMModel, *args, **kwargs
    ) -> list[LLMResult]:
        return await model.call(*args, **kwargs)

    @pytest.mark.vcr
    @pytest.mark.parametrize("model_name", ["gpt-3.5-turbo"])
    @pytest.mark.asyncio
    async def test_model(self, model_name: str) -> None:
        # Make model_name an arg so that TestLLMModel can parametrize it
        # only testing OpenAI, as other APIs don't support n>1
        model = self.MODEL_CLS(name=model_name, config=self.DEFAULT_CONFIG)
        messages = [
            Message(role="system", content="Respond with single words."),
            Message(content="Hello, how are you?"),
        ]
        results = await self.call_model(model, messages)
        assert len(results) == self.NUM_COMPLETIONS

        for result in results:
            assert result.prompt_count > 0
            assert result.completion_count > 0
            prompt_cost, completion_cost = result.prompt_and_completion_costs
            assert prompt_cost > 0
            assert completion_cost > 0
            assert result.logprob is None or result.logprob <= 0

    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, "gpt-3.5-turbo"]
    )
    @pytest.mark.asyncio
    async def test_streaming(self, model_name: str) -> None:
        model = self.MODEL_CLS(name=model_name, config=self.DEFAULT_CONFIG)
        messages = [
            Message(role="system", content="Respond with single words."),
            Message(content="Hello, how are you?"),
        ]

        def callback(_) -> None:
            return

        with pytest.raises(
            NotImplementedError,
            match="Multiple completions with callbacks is not supported",
        ):
            await self.call_model(model, messages, [callback])

    @pytest.mark.vcr
    @pytest.mark.asyncio
    async def test_parameterizing_tool_from_arg_union(self) -> None:
        def play(move: int | None) -> None:
            """Play one turn by choosing a move.

            Args:
                move: Choose an integer to lose, choose None to win.
            """

        results = await self.call_model(
            self.MODEL_CLS(name="gpt-3.5-turbo", config=self.DEFAULT_CONFIG),
            messages=[Message(content="Please win.")],
            tools=[Tool.from_function(play)],
        )
        assert len(results) == self.NUM_COMPLETIONS
        for result in results:
            assert result.messages
            assert len(result.messages) == 1
            assert isinstance(result.messages[0], ToolRequestMessage)
            assert result.messages[0].tool_calls
            assert result.messages[0].tool_calls[0].function.arguments["move"] is None

    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_output_schema(self) -> None:
        model = self.MODEL_CLS(name="gpt-3.5-turbo", config=self.DEFAULT_CONFIG)
        messages = [
            Message(
                content=(
                    "My name is Claude and I am 1 year old. What is my name and age?"
                )
            ),
        ]
        results = await self.call_model(model, messages, output_type=DummyOutputSchema)
        assert len(results) == self.NUM_COMPLETIONS
        for result in results:
            assert result.messages
            assert len(result.messages) == 1
            assert result.messages[0].content
            DummyOutputSchema.model_validate_json(result.messages[0].content)

    @pytest.mark.parametrize("model_name", [CILLMModelNames.OPENAI.value])
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_text_image_message(self, model_name: str) -> None:
        model = self.MODEL_CLS(name=model_name, config=self.DEFAULT_CONFIG)

        # An RGB image of a red square
        image = np.zeros((32, 32, 3), dtype=np.uint8)
        image[:] = [255, 0, 0]  # (255 red, 0 green, 0 blue) is maximum red in RGB

        results = await self.call_model(
            model,
            messages=[
                Message.create_message(
                    text="What color is this square? Respond only with the color name.",
                    image=image,
                )
            ],
        )
        assert len(results) == self.NUM_COMPLETIONS
        for result in results:
            assert (
                result.messages is not None
            ), "Expected messages in result, but got None"
            assert (
                result.messages[-1].content is not None
            ), "Expected content in message, but got None"
            assert "red" in result.messages[-1].content.lower()


class TestLLMModel(TestMultipleCompletionLLMModel):
    NUM_COMPLETIONS: ClassVar[int] = 1
    DEFAULT_CONFIG: ClassVar[dict] = {}
    MODEL_CLS: ClassVar[type[MultipleCompletionLLMModel]] = LLMModel

    async def call_model(self, model: LLMModel, *args, **kwargs) -> list[LLMResult]:  # type: ignore[override]
        return [await model.call(*args, **kwargs)]

    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, "gpt-3.5-turbo"]
    )
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_model(self, model_name: str) -> None:
        await super().test_model(model_name)

    @pytest.mark.vcr
    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, "gpt-3.5-turbo"]
    )
    @pytest.mark.asyncio
    async def test_streaming(self, model_name: str) -> None:
        model = LLMModel(name=model_name)
        messages = [
            Message(role="system", content="Respond with single words."),
            Message(content="Hello, how are you?"),
        ]
        content = []

        def callback(s):
            content.append(s)

        result = await model.call(messages, [callback])
        assert result.completion_count > 0
        assert content

    @pytest.mark.vcr
    @pytest.mark.asyncio
    async def test_parameterizing_tool_from_arg_union(self) -> None:
        await super().test_parameterizing_tool_from_arg_union()

    @pytest.mark.asyncio
    async def test_output_type_rejected_validation(self) -> None:
        class InstructionList(BaseModel):
            instructions: list[str] = Field(description="list of instructions")

        model = self.MODEL_CLS(name=CILLMModelNames.ANTHROPIC.value)
        with pytest.raises(
            litellm.UnsupportedParamsError,
            match="anthropic does not support parameters",
        ):
            await model.call(
                [Message(content="What are three things I should do today?")],
                output_type=InstructionList,
            )

    @pytest.mark.parametrize(
        "model_name",
        [CILLMModelNames.ANTHROPIC.value, "gpt-4-turbo", CILLMModelNames.OPENAI.value],
    )
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_text_image_message(self, model_name: str) -> None:
        await super().test_text_image_message(model_name)



from aviary.env import TASK_DATASET_REGISTRY
from aviary.env import DummyTaskDataset as _DummyTaskDataset

from .callbacks import ComputeTrajectoryMetricsMixin


class DummyTaskDataset(_DummyTaskDataset, ComputeTrajectoryMetricsMixin):
    pass


TASK_DATASET_REGISTRY["dummy"] = "ldp.alg.datasets", "DummyTaskDataset"



import itertools
from collections.abc import Sequence
from typing import Any

import networkx as nx
from aviary.message import Message, join
from aviary.tools import Tool, ToolRequestMessage

from ldp.graph import OpResult
from ldp.graph.ops import GradOutType


def to_network(  # noqa: C901
    op_result: OpResult,
    max_label_height: int | None = None,
    max_label_width: int | None = None,
    G: "nx.MultiDiGraph | None" = None,
) -> "nx.MultiDiGraph":
    """
    Populate a NetworkX graph from the input op result's computation graph.

    How to export Graphviz .dot file: nx.drawing.nx_pydot.write_dot(G, "file.dot")
    How to render with Graphviz: nx.drawing.nx_pydot.to_pydot(G).write_png("file.png")
    Online Graphviz renderer: https://dreampuf.github.io/GraphvizOnline/

    Args:
        op_result: Starting op result to recurse parent op calls and results.
        max_label_height: Optional max label height (lines).
        max_label_width: Optional max label width (chars).
        G: Optional graph to add nodes/edges to. Allows this to be a recursive function.

    Returns:
        Populated a NetworkX multi-edge directed graph.
    """

    def gvizify(x: Any) -> str:
        """Stringify and then escape colons for Graphviz labels."""
        if isinstance(x, OpResult):
            x = x.value
        if isinstance(x, Sequence):
            if isinstance(x[0], Message):
                x = join(x)
            elif isinstance(x[0], Tool):
                x = "\n".join(f"Tool {t.info.name}" for t in x)
        elif isinstance(x, ToolRequestMessage):
            # reformatting tool calls to make them easier to read
            x = str(x).split(" for tool calls: ")
            x = "\n".join(x).replace("; ", "\n")
        result = (
            "\n".join(
                # Replace double quotes since they can interfere with colon escapes
                # Strip here to avoid trailing spaces in the labels
                x_line[:max_label_width].replace('"', "'").strip()
                for i, x_line in enumerate(str(x).split("\n"))
                if not max_label_height or i < max_label_height
            )
        ).strip()  # Remove trailing newlines
        return result if ":" not in result else f'"{result}"'  # Escape colons

    call_id = op_result.call_id
    assert (
        call_id is not None
    ), "to_network currently assumes a compute graph is available"
    ctx = op_result.ctx

    op_result_str = gvizify(op_result)
    op_result_node = gvizify(f"{op_result_str}\n{call_id.fwd_id}")
    if G is None:
        # TODO: figure out a way to use OpResult.get_compute_graph(), which builds
        # a nx.DiGraph.
        G = nx.MultiDiGraph()

    op_call_str = gvizify(f"{ctx.op_name}:{call_id.fwd_id}")
    if op_call_str in G:
        # We have already visited this node - can skip.
        return G

    G.add_node(op_result_node, style="dotted", label=op_result_str)
    G.add_edge(op_call_str, op_result_node)

    if (
        result_grad := ctx.get(key="grad_output", call_id=call_id, default=None)
    ) is not None:
        G.add_edge(
            op_result_node,
            op_call_str,
            label=gvizify(result_grad),
            style="dotted",
        )

    input_args, input_kwargs = op_result.inputs
    grads = ctx.get(key="grad_input", call_id=call_id, default=None)
    if grads is None:
        arg_grads: list[GradOutType | None] = [None] * len(input_args)
        kwarg_grads: dict[str, GradOutType | None] = dict.fromkeys(input_kwargs)
    else:
        arg_grads, kwarg_grads = grads

    args_and_grads = itertools.chain(
        zip(input_args, arg_grads, strict=True),
        ((arg, kwarg_grads[key]) for key, arg in input_kwargs.items()),
    )

    for arg, grad in args_and_grads:
        arg_str = gvizify(arg)

        if not isinstance(arg, OpResult):
            G.add_node(arg_str, style="dotted")

        else:
            arg_str = gvizify(f"{arg_str}\n{arg.call_id.fwd_id}")
            G = to_network(
                arg,
                max_label_height=max_label_height,
                max_label_width=max_label_width,
                G=G,
            )

        G.add_edge(arg_str, op_call_str)
        if grad is not None:
            G.add_edge(op_call_str, arg_str, label=gvizify(grad), style="dotted")

    return G



import asyncio
import itertools
import logging
import uuid
from collections.abc import Callable, Iterator, Sequence
from contextlib import contextmanager, nullcontext
from typing import Any, TypeVar, overload

from aviary.env import Environment
from aviary.message import Message

from ldp.agent import Agent
from ldp.data_structures import Trajectory, Transition

from .callbacks import Callback

logger = logging.getLogger(__name__)


TEnv = TypeVar("TEnv", bound=Environment)


class CaughtError(Exception):
    """Base class for reraised exceptions when catching is enabled."""

    def __init__(self, original_exc: Exception):
        self.original_exc = original_exc

    exc_type = "undefined"


class AgentError(CaughtError):
    exc_type = "agent"


class EnvError(CaughtError):
    exc_type = "env"


@contextmanager
def reraise_exc_as(reraise: type[CaughtError], enabled: bool) -> Iterator[None]:
    try:
        yield
    except Exception as e:
        if enabled:
            logging.exception(f"Caught {reraise.exc_type} exception.")
            raise reraise(e) from e
        raise


class RolloutManager:
    def __init__(
        self,
        agent: Agent,
        catch_agent_failures: bool = True,
        catch_env_failures: bool = True,
        callbacks: Sequence[Callback] | None = None,
        concurrency_limit: int | None = None,
    ):
        self.agent = agent

        self.catch_agent_failures = catch_agent_failures
        self.catch_env_failures = catch_env_failures

        self.concurrency_limiter = (
            asyncio.Semaphore(concurrency_limit) if concurrency_limit else nullcontext()
        )

        self.traj_buffer: dict[str, Trajectory] = {}
        self.callbacks = callbacks or []

    @overload
    async def sample_trajectories(  # noqa: D418
        self,
        environment_factory: Callable[[], TEnv],
        batch_size: int = 1,
        max_steps: int | None = None,
    ) -> list[tuple[Trajectory, TEnv]]:
        """Run rollouts in parallel, using a factory to construct environments.

        We will construct `batch_size` environments and run rollouts on each of them.
        If `max_steps` is set, rollouts will be truncated at this value. If a rollout
        has fewer than `max_steps`, then a new environment will be constructed and another
        rollout will be started until `max_steps` is reached.

        Args:
            environment_factory: A no-argument callable that returns
                an environment instance
            batch_size (int, optional): Defaults to 1.
            max_steps (int | None, optional): Max steps per rollout. Defaults to None (see above).

        Returns:
            list[tuple[Trajectory, Environment]]: A list of (trajectory, environment) tuples: one per rollout.
        """

    @overload
    async def sample_trajectories(  # noqa: D418
        self,
        environments: Sequence[Environment],
        max_steps: int | None = None,
    ) -> list[Trajectory]:
        """Run rollouts in parallel on a list of provided environments.

        Args:
            environments: A list of environments to run rollouts on.
            max_steps: Max steps per rollout. Defaults to None, in which case the rollouts are run
                until environment returns done.
        """

    async def sample_trajectories(self, **kwargs):
        if "environment_factory" in kwargs:
            assert (
                "environments" not in kwargs
            ), "Cannot use environment_factory with environments"

            return await self._sample_trajectories_from_env_factory(
                kwargs["environment_factory"],
                kwargs.get("batch_size", 1),
                kwargs.get("max_steps"),
            )

        if "environments" in kwargs:
            assert (
                "environment_factory" not in kwargs
            ), "Cannot use environments with environment_factory"
            return await self._sample_trajectories_from_envs(
                kwargs["environments"], kwargs.get("max_steps")
            )

        raise TypeError(
            "sample_trajectories() missing required "
            "arguments 'environment_factory' or 'environments'"
        )

    async def _sample_trajectories_from_env_factory(
        self,
        environment_factory: Callable[[], Environment],
        batch_size: int = 1,
        max_steps: int | None = None,
    ) -> list[tuple[Trajectory, Environment]]:
        self.traj_buffer.clear()

        async def rollout_with_args(idx: int, **rollout_kwargs):
            return idx, await self._rollout(**rollout_kwargs), rollout_kwargs

        accumulated_steps = [0] * batch_size
        # submit initial batch of tasks
        tasks = [
            asyncio.create_task(
                rollout_with_args(
                    idx,
                    traj_id=uuid.uuid4().hex,
                    env=environment_factory(),
                    max_steps=max_steps,
                )
            )
            for idx in range(batch_size)
        ]

        results = []
        while tasks:
            done, pending = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED
            )
            new_tasks = []
            for task in done:
                idx, traj, kwargs = await task
                results.append((traj, kwargs["env"]))
                accumulated_steps[idx] += len(traj.steps)
                if (
                    max_steps is not None
                    and (remaining_steps := max_steps - accumulated_steps[idx]) > 0
                ):
                    # submit another task if we haven't reached max_steps
                    new_task = asyncio.create_task(
                        rollout_with_args(
                            idx,
                            traj_id=uuid.uuid4().hex,
                            env=environment_factory(),
                            max_steps=remaining_steps,
                        )
                    )
                    new_tasks.append(new_task)

            tasks = list(pending) + new_tasks

        return results

    async def _sample_trajectories_from_envs(
        self,
        environments: Sequence[Environment],
        max_steps: int | None = None,
    ) -> list[Trajectory]:
        self.traj_buffer.clear()

        traj_ids = [uuid.uuid4().hex for _ in range(len(environments))]
        await asyncio.gather(
            *(
                self._rollout(*args, max_steps=max_steps)
                for args in zip(traj_ids, environments, strict=True)
            )
        )
        return [self.traj_buffer[traj_id] for traj_id in traj_ids]

    async def _rollout(
        self,
        traj_id: str,
        env: Environment,
        max_steps: int | None,
    ) -> Trajectory:
        trajectory = Trajectory(traj_id=traj_id)

        async def store_step(step: Transition):
            await asyncio.gather(*[
                callback.after_transition(traj_id, self.agent, env, step)
                for callback in self.callbacks
            ])
            trajectory.steps.append(step)

        # Set default values to store in the buffer in case reset/init_state fail
        obs: list[Message] = []
        agent_state: Any = None

        try:
            with reraise_exc_as(EnvError, enabled=self.catch_env_failures):
                obs, tools = await env.reset()
            await asyncio.gather(*[
                c.after_env_reset(traj_id, obs, tools) for c in self.callbacks
            ])

            with reraise_exc_as(AgentError, enabled=self.catch_agent_failures):
                agent_state = await self.agent.init_state(tools)
            await asyncio.gather(*[
                c.after_agent_init_state(traj_id, agent_state) for c in self.callbacks
            ])

            for timestep in itertools.count():
                step = await self._take_step(timestep, traj_id, env, agent_state, obs)

                if timestep + 1 == max_steps and not step.done:
                    # Mark as truncated if we hit max_steps and the state is not terminal.
                    # Do it before store_step(), so that callbacks can access this info
                    step.truncated = True

                # We assume the below won't throw a CaughtError
                await store_step(step)

                # set things up for the next iteration
                agent_state = step.next_agent_state
                obs = step.next_observation

                if step.done or step.truncated:
                    break

        except CaughtError as e:
            # NOTE: This trajectory should not be used for regular training.
            # We save the last transition here for debugging, etc.
            await store_step(
                Transition(
                    timestep=len(trajectory.steps),
                    agent_state=agent_state,
                    next_agent_state=None,
                    observation=obs,
                    next_observation=[],
                    action=None,
                    done=True,
                    metadata={"exception": repr(e.original_exc)},
                )
            )

        self.traj_buffer[traj_id] = trajectory
        return trajectory

    async def _take_step(
        self,
        timestep: int,
        traj_id: str,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
    ) -> Transition:
        async with self.concurrency_limiter:
            await asyncio.gather(*[
                callback.before_transition(traj_id, self.agent, env, agent_state, obs)
                for callback in self.callbacks
            ])

            with reraise_exc_as(AgentError, enabled=self.catch_agent_failures):
                (
                    action,
                    next_agent_state,
                    value,
                ) = await self.agent.get_asv(agent_state, obs)
            await asyncio.gather(*[
                callback.after_agent_get_asv(traj_id, action, next_agent_state, value)
                for callback in self.callbacks
            ])

            with reraise_exc_as(EnvError, enabled=self.catch_env_failures):
                next_obs, reward, done, trunc = await env.step(action.value)
            await asyncio.gather(*[
                callback.after_env_step(traj_id, next_obs, reward, done, trunc)
                for callback in self.callbacks
            ])

            return Transition(
                timestep=timestep,
                agent_state=agent_state,
                next_agent_state=next_agent_state,
                action=action,
                reward=reward,
                value=value,
                observation=obs,
                next_observation=next_obs,
                done=done,
                truncated=trunc,
            )



import itertools
import random
import tempfile
from copy import deepcopy
from typing import Any, cast

import pytest
from aviary.env import Environment, Frame
from aviary.message import Message
from aviary.tools import Tool, ToolRequestMessage
from pydantic import BaseModel

from ldp.agent import Agent, SimpleAgent, SimpleAgentState
from ldp.alg import BeamSearchRollout, Callback, RolloutManager, TreeSearchRollout
from ldp.data_structures import Trajectory, Transition
from ldp.graph import FxnOp, OpResult, compute_graph, set_training_mode


class DummyEnv(Environment[None]):
    def __init__(self):
        self.tools = [Tool.from_function(self.talk)]

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        return [Message(content="Hello!")], self.tools

    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[list[Message], float, bool, bool]:
        if action.tool_calls:
            responses = cast(list[Message], await self.exec_tool_calls(action))
        else:
            responses = [Message(content="Use the 'talk' tool to speak.")]

        return responses, 0.0, False, False

    async def talk(self, message: str) -> str:
        """Say something to me.

        Args:
            message (str): what you want to say

        Returns:
            str: my response
        """
        return message

    def export_frame(self) -> Frame:
        return Frame()


async def count_exclamations(traj: Trajectory) -> float:
    last_step = traj.steps[-1]
    agent_state = cast(SimpleAgentState, last_step.next_agent_state)
    return float(
        sum(m.content.count("!") for m in agent_state.messages if m.content is not None)
    )


@pytest.mark.vcr
@pytest.mark.parametrize("training", [True, False])
@pytest.mark.asyncio
async def test_rollout(training: bool) -> None:
    agent = SimpleAgent()
    callback = DummyCallback()
    set_training_mode(training)
    rollout_manager = RolloutManager(
        agent,
        catch_agent_failures=False,
        catch_env_failures=False,
        callbacks=[callback],
    )
    trajs = await rollout_manager.sample_trajectories(
        environments=[DummyEnv(), DummyEnv()], max_steps=1
    )
    assert len(trajs) == 2

    # Let's check we can serialize and deserialize the trajectories
    for traj in trajs:
        with tempfile.NamedTemporaryFile(suffix=".jsonl") as f:
            traj.to_jsonl(filename=f.name)
            rehydrated_traj = Trajectory.from_jsonl(f.name)
            assert traj.traj_id == rehydrated_traj.traj_id

    assert all(v == 2 for v in callback.fn_invocations.values())


async def adeepcopy(x):
    return deepcopy(x)


@pytest.mark.vcr
@pytest.mark.asyncio
async def test_beam_search() -> None:
    agent = SimpleAgent()
    callback = DummyCallback()
    beam_search = BeamSearchRollout(
        agent,
        beam_width=1,  # keep these numbers small to speed up test
        samples_per_beam=1,
        env_clone_fn=adeepcopy,
        agent_clone_fn=deepcopy,
        scoring_fn=count_exclamations,
        catch_agent_failures=False,
        catch_env_failures=False,
        callbacks=[callback],
    )

    trajs = await beam_search.sample_trajectories(
        environments=[DummyEnv(), DummyEnv()], max_steps=1
    )
    assert len(trajs) == 2

    assert all(
        v == 2
        for k, v in callback.fn_invocations.items()
        if k != "after_agent_init_state"  # TODO: support this callback too
    )


class DummyCallback(Callback):
    def __init__(self):
        # NOTE: don't use collections.defaultdict here because it can lead to
        # test aliasing for a callback being missed altogether
        self.fn_invocations = {
            "before_transition": 0,
            "after_agent_init_state": 0,
            "after_agent_get_asv": 0,
            "after_env_reset": 0,
            "after_env_step": 0,
            "after_transition": 0,
        }

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
    ) -> None:
        self.fn_invocations["before_transition"] += 1

    async def after_agent_init_state(self, traj_id: str, init_state: Any) -> None:
        self.fn_invocations["after_agent_init_state"] += 1

    async def after_agent_get_asv(
        self,
        traj_id: str,
        action: OpResult[ToolRequestMessage],
        next_agent_state: Any,
        value: float,
    ):
        self.fn_invocations["after_agent_get_asv"] += 1

    async def after_env_reset(
        self, traj_id: str, obs: list[Message], tools: list[Tool]
    ) -> None:
        self.fn_invocations["after_env_reset"] += 1

    async def after_env_step(
        self,
        traj_id: str,
        obs: list[Message],
        reward: float,
        done: bool,
        trunc: bool,
    ):
        self.fn_invocations["after_env_step"] += 1

    async def after_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        transition: Transition,
    ) -> None:
        self.fn_invocations["after_transition"] += 1


class CountingAgentState(BaseModel):
    count: float = 0.0


class CountingAgent(Agent[CountingAgentState]):
    def __init__(self):
        self.op = FxnOp[ToolRequestMessage](lambda: ToolRequestMessage(tool_calls=[]))

    async def init_state(self, tools: list[Tool]) -> CountingAgentState:
        return CountingAgentState()

    @compute_graph()
    async def get_asv(
        self, agent_state: CountingAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], CountingAgentState, float]:
        new_state = CountingAgentState(count=float(cast(str, obs[0].content)) + 1)
        action = await self.op()
        return action, new_state, 0.0


class CountingEnv(Environment[float]):
    def __init__(self, state: float = 0.0):
        self.state = state

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        return [Message(content=str(self.state))], []

    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[list[Message], float, bool, bool]:
        self.state += 1
        return [Message(content=str(self.state))], 0.0, self.state >= 3, False

    def export_frame(self) -> Frame:
        return Frame()


@pytest.mark.asyncio
async def test_deterministic_rollout():
    agent = CountingAgent()
    env = CountingEnv()

    rollout_manager = RolloutManager(agent)
    traj, *_ = await rollout_manager.sample_trajectories(environments=[env])

    assert len(traj.steps) == 3
    for i_step, step in enumerate(traj.steps):
        f_step = float(i_step)
        # check that we didn't clobber any agent or env states
        assert step.agent_state.count == f_step
        assert step.next_agent_state.count == f_step + 1
        assert step.observation[0].content == str(f_step)
        assert step.next_observation[0].content == str(f_step + 1)


class NoisyCountingEnv(CountingEnv):
    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[list[Message], float, bool, bool]:
        self.state += 1 + random.uniform(-0.01, 0.01)
        return [Message(content=str(self.state))], 1.0, self.state >= 3, False


class TestTreeSearch:
    @pytest.mark.asyncio
    async def test_tree_search(self) -> None:
        agent = CountingAgent()
        # Use a slightly stochastic env so we can distinguish branches
        env = NoisyCountingEnv()

        callback = DummyCallback()
        rollout_manager = TreeSearchRollout(
            agent,
            branching_factor=2,
            env_clone_fn=deepcopy,
            concurrency_limit=1,
            callbacks=[callback],
        )
        tree = await rollout_manager.sample_tree(env, max_depth=3)
        trajs = tree.get_trajectories()
        assert len(trajs) == 8

        traj_ids_wo_root: set[str] = {
            cast(str, traj.traj_id).replace(tree.root_id, "").lstrip(":")
            for traj in trajs
        }
        # IDs should be 0:0:0, 0:0:1, ... 1:1:1 (order doesn't matter)
        assert traj_ids_wo_root == {
            ":".join(x) for x in itertools.product("01", repeat=3)
        }

        observations: dict[tuple[str, ...], str] = {}
        for traj in trajs:
            branch_path = tuple(cast(str, traj.traj_id).split(":")[1:])

            prev_step: Transition | None = None
            for i_step, step in enumerate(traj.steps):
                if prev_step is not None:
                    # Check that the child node started at the state emitted at the parent node
                    assert prev_step.next_agent_state == step.agent_state

                # Steps that started at the same node in the tree should have the same observation
                node_id = branch_path[: i_step + 1]
                assert len(step.observation) == 1
                assert step.observation[0].content
                if node_id in observations:
                    assert observations[node_id] == step.observation[0].content
                else:
                    observations[node_id] = step.observation[0].content

                prev_step = step

        for callback_fn, num_calls in callback.fn_invocations.items():
            if callback_fn in {"after_agent_init_state", "after_env_reset"}:
                assert num_calls == 1, "These should be invoked once at the start"
            else:
                # We expect sum_{i=1}^3 2^i = 2^4 - 2 = 14 transitions:
                # - branching factor = 2, depth = 3
                # - root node isn't sampled, so no i=0 term in sum
                assert num_calls == 14

    @pytest.mark.asyncio
    async def test_early_stopping(self) -> None:
        agent = CountingAgent()
        # Use a slightly stochastic env so we can distinguish branches
        env = NoisyCountingEnv()

        callback = DummyCallback()
        rollout_manager = TreeSearchRollout(
            agent,
            branching_factor=2,
            env_clone_fn=deepcopy,
            concurrency_limit=1,
            callbacks=[callback],
            target_reward=0.5,
        )
        trajs = (await rollout_manager.sample_tree(env, max_depth=3)).get_trajectories()
        assert len(trajs) < 8  # should have exited early
        for traj in trajs:
            # should have hit target reward immediately
            assert len(traj.steps) == 1



import contextvars
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from uuid import UUID, uuid4

from aviary.utils import is_coroutine_callable
from pydantic import BaseModel, field_serializer, field_validator


class CallID(BaseModel):
    run_id: UUID
    fwd_id: UUID

    def __repr__(self) -> str:
        return f"{self.run_id}:{self.fwd_id}"

    def __hash__(self) -> int:
        return hash((self.run_id, self.fwd_id))

    @field_validator("run_id", "fwd_id", mode="before")
    @classmethod
    def validate_uuid(cls, value: UUID | str) -> UUID:
        if isinstance(value, str):
            return UUID(value)
        return value

    @field_serializer("run_id", "fwd_id")
    def serialize_uuid(self, value: UUID) -> str:
        return str(value)

    def __init__(self, run_id: str | UUID, fwd_id: str | UUID):
        # Convenience so we can use positional arguments
        super().__init__(run_id=run_id, fwd_id=fwd_id)


_RUN_ID = contextvars.ContextVar[UUID]("run_id")
_CALL_ID = contextvars.ContextVar[CallID]("call_id")


@asynccontextmanager
async def compute_graph() -> AsyncIterator[UUID]:
    """Initialize a compute graph by setting a run ID.

    If a run ID is already set (i.e. we are already inside a
    get_run_id() context), then the existing run ID is returned.
    Otherwise, a new UUID is created.
    """
    try:
        # If a run ID is set, return it.
        run_id = _RUN_ID.get()
        token: contextvars.Token | None = None
    except LookupError:
        # If not, make a new run ID.
        run_id = uuid4()
        token = _RUN_ID.set(run_id)

    try:
        yield run_id
    finally:
        if token is not None:
            # token is not None if we made a new run ID. In that case,
            # reset the context to its previous state.
            _RUN_ID.reset(token)


def get_run_id() -> UUID:
    """Get the current run ID."""
    try:
        return _RUN_ID.get()
    except LookupError:
        raise RuntimeError(
            "Attempting to access run ID, but not inside compute graph context."
        ) from None


@asynccontextmanager
async def op_call() -> AsyncIterator[CallID]:
    """Decorate an op call with a call ID.

    If a call ID is already set (i.e. we are already inside an op call),
    then the existing call ID is returned.
    Otherwise, a new UUID is created.
    """
    # Get run_id in case we need to construct a CallID, but this also serves
    # as a check that we're inside compute_graph()
    run_id = get_run_id()

    try:
        call_id = _CALL_ID.get()
        token: contextvars.Token | None = None
    except LookupError:
        fwd_id = uuid4()
        call_id = CallID(run_id, fwd_id)
        token = _CALL_ID.set(call_id)

    try:
        yield call_id
    finally:
        if token is not None:
            # token is not None if we made a new call ID. In that case,
            # reset the context to its previous state.
            _CALL_ID.reset(token)


def get_call_id() -> CallID:
    """Get the current call ID."""
    try:
        return _CALL_ID.get()
    except LookupError:
        raise RuntimeError(
            "Attempting to access call ID, but not inside op call context."
        ) from None


_TRAINING_MODE = contextvars.ContextVar[bool]("training_mode", default=True)


def get_training_mode() -> bool:
    """Get the current training mode."""
    return _TRAINING_MODE.get()


def set_training_mode(training_mode: bool) -> None:
    """Set the training mode."""
    _TRAINING_MODE.set(training_mode)


class _TrainingModeContext:
    """Automatically set and reset the training_mode with a context manager."""

    def __init__(self, training_mode: bool):
        self.training_mode = training_mode
        self.prev_training_mode = get_training_mode()

    def __call__(self, fn=None):
        if fn is None:
            return self

        if is_coroutine_callable(fn):

            async def wrapper(*args, **kwargs):
                async with self:
                    return await fn(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                with self:
                    return fn(*args, **kwargs)

        return wrapper

    def __enter__(self) -> None:
        self.prev_training_mode = get_training_mode()
        set_training_mode(self.training_mode)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        set_training_mode(self.prev_training_mode)

    async def __aenter__(self) -> None:
        self.__enter__()

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        self.__exit__(exc_type, exc_value, traceback)


train_mode = _TrainingModeContext(training_mode=True)
eval_mode = _TrainingModeContext(training_mode=False)



import pytest
from aviary.env import DummyEnv

from ldp.agent import SimpleAgent
from ldp.utils import discounted_returns


@pytest.mark.asyncio
async def test_rollout_and_discounting(dummy_env: DummyEnv) -> None:
    obs, tools = await dummy_env.reset()

    agent = SimpleAgent(tools=tools)
    agent_state = await agent.init_state(tools=tools)

    observations = []
    actions = []
    rewards = []
    terms = []
    done = True
    for i in range(3):  # noqa: B007
        if done:
            obs, _ = await dummy_env.reset()
            agent_state = await agent.init_state(tools=tools)

        observations.append((obs, agent_state))
        action, agent_state, _ = await agent.get_asv(agent_state, obs)
        obs, reward, done, _ = await dummy_env.step(action.value)
        actions.append(action)
        rewards.append(reward)
        terms.append(done)

    print(terms)
    d_returns = discounted_returns(rewards, terms, 0.5)
    print(d_returns)



from enum import StrEnum


class DefaultLLMModelNames(StrEnum):
    """Defaults for LLM models, pin exact versions for performance stability."""

    OPENAI = "gpt-4o-2024-08-06"  # Cheap, fast, and decent


# ruff: noqa: E402  # Avoid circular imports

from .agent import Agent, AgentConfig
from .agent_client import HTTPAgentClient, make_simple_agent_server
from .memory_agent import MemoryAgent
from .react_agent import ReActAgent
from .simple_agent import SimpleAgent, SimpleAgentState
from .tree_of_thoughts_agent import TreeofThoughtsAgent

__all__ = [
    "Agent",
    "AgentConfig",
    "DefaultLLMModelNames",
    "HTTPAgentClient",
    "MemoryAgent",
    "ReActAgent",
    "SimpleAgent",
    "SimpleAgentState",
    "TreeofThoughtsAgent",
    "make_simple_agent_server",
]



import asyncio
from unittest.mock import MagicMock, patch

import litellm
import numpy as np
import pytest
from litellm.caching import Cache, InMemoryCache
from pytest_subtests import SubTests

from ldp.llms import (
    EmbeddingModel,
    HybridEmbeddingModel,
    LiteEmbeddingModel,
    SparseEmbeddingModel,
)


class TestLiteEmbeddingModel:
    @pytest.mark.asyncio
    async def test_embed_texts(self) -> None:
        texts = ["Hello", "World"]
        batch_size = 1  # NOTE: this affects the mock below
        model = LiteEmbeddingModel(name="stub", batch_size=1)
        with patch(
            "litellm.aembedding",
            autospec=True,
            side_effect=[
                MagicMock(data=[{"embedding": [1.0, 2.0]}]),
                MagicMock(data=[{"embedding": [3.0, 4.0]}]),
            ],
        ) as mock_aembedding:
            embeddings = await model.embed_texts(texts)

        assert np.allclose(embeddings[0], [1.0, 2.0])
        assert np.allclose(embeddings[1], [3.0, 4.0])
        assert mock_aembedding.call_count == len(texts) / batch_size

    @pytest.mark.parametrize(
        ("model_name", "expected_dimensions"),
        [
            ("stub", None),
            ("text-embedding-ada-002", 1536),
            ("text-embedding-3-small", 1536),
        ],
    )
    def test_model_dimension_inference(
        self, model_name: str, expected_dimensions: int | None
    ) -> None:
        assert LiteEmbeddingModel(name=model_name).dimensions == expected_dimensions

    @pytest.mark.asyncio
    async def test_can_change_dimension(self) -> None:
        """We run this one for real, because want to test end to end."""
        stub_texts = ["test1", "test2"]

        model = LiteEmbeddingModel(name="text-embedding-3-small")
        assert model.dimensions == 1536

        model = LiteEmbeddingModel(name="text-embedding-3-small", dimensions=8)
        assert model.dimensions == 8
        etext1, etext2 = await model.embed_texts(stub_texts)
        assert len(etext1) == len(etext2) == 8

    @pytest.mark.vcr
    @pytest.mark.asyncio
    async def test_caching(self) -> None:
        model = LiteEmbeddingModel(
            name="text-embedding-3-small", dimensions=8, embed_kwargs={"caching": True}
        )
        # Make sure there is no existing cache.
        with patch("litellm.cache", None):
            # now create a new cache
            litellm.cache = Cache()
            assert isinstance(litellm.cache.cache, InMemoryCache)
            assert len(litellm.cache.cache.cache_dict) == 0

            _ = await model.embed_texts(["test1"])
            # need to do this to see the data propagated to cache
            await asyncio.sleep(0.0)

            # Check the cache entry was made
            assert len(litellm.cache.cache.cache_dict) == 1


@pytest.mark.asyncio
async def test_sparse_embedding_model(subtests: SubTests):
    with subtests.test("1D sparse"):
        ndim = 1
        expected_output = [[1.0], [1.0]]

        model = SparseEmbeddingModel(dimensions=ndim)
        result = await model.embed_texts(["test1", "test2"])

        assert result == expected_output

    with subtests.test("large sparse"):
        ndim = 1024

        model = SparseEmbeddingModel(dimensions=ndim)
        result = await model.embed_texts(["hello test", "go hello"])

        assert max(result[0]) == max(result[1]) == 0.5

    with subtests.test("default sparse"):
        model = SparseEmbeddingModel()
        result = await model.embed_texts(["test1 hello", "test2 hello"])

        assert pytest.approx(sum(result[0]), abs=1e-6) == pytest.approx(
            sum(result[1]), abs=1e-6
        )


@pytest.mark.asyncio
async def test_hybrid_embedding_model() -> None:
    hybrid_model = HybridEmbeddingModel(
        models=[LiteEmbeddingModel(), SparseEmbeddingModel()]
    )

    # Mock the embedded documents of Lite and Sparse models
    with (
        patch.object(LiteEmbeddingModel, "embed_texts", return_value=[[1.0], [2.0]]),
        patch.object(SparseEmbeddingModel, "embed_texts", return_value=[[3.0], [4.0]]),
    ):
        result = await hybrid_model.embed_texts(["hello", "world"])
    assert result.tolist() == [[1.0, 3.0], [2.0, 4.0]]


@pytest.mark.asyncio
async def test_class_constructor() -> None:
    original_name = "hybrid-text-embedding-3-small"
    model = EmbeddingModel.from_name(original_name)
    assert isinstance(model, HybridEmbeddingModel)
    assert model.name == original_name
    dense_model, sparse_model = model.models
    assert dense_model.name == "text-embedding-3-small"
    assert dense_model.dimensions == 1536
    assert sparse_model.name == "sparse"
    assert sparse_model.dimensions == 256
    assert model.dimensions == 1792



import logging
from typing import Any, Self, cast

from aviary.message import MalformedMessageError, Message
from aviary.tools import Tool, ToolRequestMessage, ToolResponseMessage
from pydantic import BaseModel, ConfigDict, Field
from tenacity import (
    Future,
    RetryCallState,
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
)

from ldp.graph import OpResult, compute_graph
from ldp.graph.modules.react import (
    ACT_DEFAULT_PROMPT_TEMPLATE,
    REACT_DEFAULT_PROMPT_TEMPLATE,
    ReActModule,
    ToolDescriptionMethods,
)

from . import DefaultLLMModelNames
from .agent import Agent
from .simple_agent import SimpleAgentState

logger = logging.getLogger(__name__)


class ReActAgent(BaseModel, Agent[SimpleAgentState]):
    """An Act or ReAct Agent built to work with chat models.

    Paper: https://arxiv.org/abs/2210.03629

    The ReAct style is like so, and note Act style has no 'Thought: ' entries:
    System:
        Answer the following questions as best you can. You have access to the following tools:

            {tools}

            Use the following format:

            Thought: you should always think about what to do
            Action: the action to take, should be one of [{tool_names}]
            Action Input: the input to the action
            Observation: the result of the action
            ... (this Thought/Action/Action Input/Observation can repeat N times)
    User:
        {questions}
    Assistant:
        Thought:
        Action:
        Action Input:
    User:
        Observation:
    Assistant:
        Thought:
        Action:
        Action Input:
    ...

    One notable design decision is that ReAct's state does not necessarily
    track ToolRequestMessage. Recall that aviary is in a partially observable
    domain, meaning we don't need to have perfect symmetry with Environments.
    Instead, ReActAgent's state stores a ReAct-style message history, where the
    messages are plain Message (and not a ToolRequestMessage).
    """

    # Freeze to ensure the only mutation happens in either the agent state (which is
    # passed around) or in the internal Ops
    model_config = ConfigDict(frozen=True)

    llm_model: dict[str, Any] = Field(
        default={
            "model": DefaultLLMModelNames.OPENAI.value,
            "temperature": 0.1,
            "logprobs": True,
            "top_logprobs": 1,
        },
        description="Starting configuration for the LLM model.",
    )
    sys_prompt: str = Field(
        default=REACT_DEFAULT_PROMPT_TEMPLATE,
        description="Learnable system prompt template, defaults to ReAct.",
    )
    tool_description_method: ToolDescriptionMethods = Field(
        default=ToolDescriptionMethods.STR,
        description="Method used to describe the tools, defaults to 'str' description.",
    )

    @classmethod
    def make_act_agent(cls, **kwargs) -> Self:
        return cls(sys_prompt=ACT_DEFAULT_PROMPT_TEMPLATE, **kwargs)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._react_module = ReActModule(
            self.llm_model, self.sys_prompt, self.tool_description_method
        )

    async def init_state(self, tools: list[Tool]) -> SimpleAgentState:
        return SimpleAgentState(tools=tools)

    @staticmethod
    def after_retry_failure_log(retry_state: RetryCallState):
        logger.error(
            f"Failed across {retry_state.attempt_number} attempts to run get_asv given"
            f" arguments {retry_state.args} and kwargs {retry_state.kwargs}."
        )
        # NOTE: this blows up with the underlying exception... it isn't wrapped in a
        # RetryError like normal tenacity
        return cast(Future, retry_state.outcome).result()

    @retry(
        retry=retry_if_exception_type(MalformedMessageError),
        before_sleep=before_sleep_log(logger, logging.WARNING),
        stop=stop_after_attempt(5),
        retry_error_callback=after_retry_failure_log,
    )
    @compute_graph()
    async def get_asv(
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        next_state = agent_state.get_next_state(
            obs=[
                Message(content=f"Observation: {m.content}")
                if isinstance(m, ToolResponseMessage)
                else m
                for m in obs
            ]
        )

        final_result, react_message = await self._react_module(
            messages=next_state.messages, tools=next_state.tools
        )
        next_state.messages = [*next_state.messages, react_message]
        return final_result, next_state, 0.0



"""Module for the Tree of Thoughts agent.

This module defines the Tree of Thoughts agent which uses a language model to generate and evaluate possible
steps in a puzzle or problem-solving environment. The agent employs a tree search mechanism to explore different
solutions and selects the most promising ones based on evaluations.

This module is based on the following paper: https://openreview.net/forum?id=5Xc1ecxO1h

Note: TreeofThoughtsAgent is currently tested as a baseline agent for Game of 24. It does not yet support tool calls
that operate on the intermediate reasoning steps. This would probably entail a redefinition of the POMDP to
undertake intermediate reasoning steps as environment steps.
"""

import logging
from collections.abc import Callable
from typing import Any

from aviary.message import Message
from aviary.tools import Tool, ToolCall, ToolRequestMessage
from pydantic import BaseModel, ConfigDict, Field

from ldp.graph import FxnOp, LLMCallOp, OpResult, compute_graph, get_call_id, op_call
from ldp.llms import prepend_sys

from . import DefaultLLMModelNames
from .agent import Agent
from .simple_agent import SimpleAgentState

logger = logging.getLogger(__name__)


class TreeofThoughtsAgent(BaseModel, Agent[SimpleAgentState]):
    """Tree of Thoughts Agent.

    This agent uses a tree search mechanism combined with an LLM to generate and evaluate
    possible steps in a problem-solving environment. It is designed to explore different solutions
    and select the most promising ones based on a heuristic evaluation function.
    """

    # Freeze to ensure the only mutation happens in either the agent state (which is
    # passed around) or in the internal Ops
    model_config = ConfigDict(frozen=True)

    llm_model: dict[str, Any] = Field(
        default={"model": DefaultLLMModelNames.OPENAI.value, "temperature": 0.1},
        description="Starting configuration for the LLM model.",
    )
    value_prompt_func: Callable[[str, str], str] = Field(
        default=lambda x, y: f"Value prompt for input: {x}, current path: {y}",
        description="Function to format value prompt template.",
    )
    proposal_prompt_func: Callable[[str, str], str] = Field(
        default=lambda x, y: f"Proposal prompt for input: {x}, current path: {y}",
        description="Function to format proposal prompt template.",
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._prepend_op = FxnOp(prepend_sys)
        self._llm_call_op = LLMCallOp()

    async def init_state(self, tools: list[Tool]) -> SimpleAgentState:
        return SimpleAgentState(tools=tools)

    @compute_graph()
    async def get_asv(  # type: ignore[override]
        self,
        agent_state: SimpleAgentState,
        obs: list[Message],
        eval_function: Callable[[str, list[str]], float],
        n_steps: int = 0,
        n_select_samples: int = 0,
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        """Generate and evaluate possible steps in the problem-solving process.

        Args:
            agent_state: The current state of the agent.
            obs: The observations provided to the agent.
            eval_function: Function to evaluate the generated paths in the tree.
            n_steps: Number of steps to generate. Defaults to 0. Dictated by the environment.
            n_select_samples: Number of tree nodes to select to explore in each step. Defaults to 0.

        Returns:
            The result of the operation, the new state of the agent, and the number representing the value (0).
        """
        new_state = agent_state.get_next_state()

        x = str(obs[0].content)  # Current problem input
        current_paths = [""]  # current candidate paths through the tree

        for step in range(n_steps):
            logger.info(f"Step {step}")

            # propose candidate paths
            candidate_paths = []
            for path in current_paths:
                proposal_prompt_init = self.proposal_prompt_func(x, path)
                proposal_msgs = await self._prepend_op(
                    new_state.messages, sys_content=proposal_prompt_init
                )
                proposal = await self._llm_call_op(self.llm_model, msgs=proposal_msgs)
                # Append candidate paths to the current paths
                candidate_paths += [
                    path + _ + "\n"
                    for _ in (proposal.value.content or "").split("\n")
                    if _
                ]

            # score candidate paths
            values = []
            for path in candidate_paths:
                value_prompt_init = self.value_prompt_func(x, path)
                value_msgs = await self._prepend_op(
                    new_state.messages, sys_content=value_prompt_init
                )
                value_outputs = await self._llm_call_op(self.llm_model, msgs=value_msgs)
                values.append(eval_function(path, [value_outputs.value.content or ""]))

            # greedy selection
            values_with_index = [(v, i) for i, v in enumerate(values)]
            sorted_values = sorted(values_with_index, key=lambda x: x[0], reverse=True)
            select_ids = [i for _, i in sorted_values[:n_select_samples]]
            select_new_paths = [candidate_paths[select_id] for select_id in select_ids]
            current_paths = select_new_paths

        # Generate tool calls for the selected answer
        tool_calls = [
            ToolCall.from_tool(tool, *[current_paths[0]]) for tool in new_state.tools
        ]
        result = ToolRequestMessage(content=current_paths[0], tool_calls=tool_calls)

        new_state.messages = [*new_state.messages, result]
        async with op_call():
            op_result: OpResult[ToolRequestMessage] = OpResult(
                call_id=get_call_id(),
                op_name="TreeofThoughtsAgentOp",
                op_class_name=type(self).__name__,
                value=result,
            )
        return op_result, new_state, 0.0



import contextlib
import shutil
import tempfile
from enum import IntEnum, auto
from functools import partial
from pathlib import Path
from unittest.mock import patch

import networkx as nx
import pytest
from aviary.env import DummyEnv
from aviary.message import Message
from aviary.tools import Tool, ToolCall, ToolRequestMessage
from httpx import ASGITransport, AsyncClient
from pydantic import BaseModel, Field

from ldp.agent import (
    Agent,
    AgentConfig,
    HTTPAgentClient,
    MemoryAgent,
    ReActAgent,
    SimpleAgent,
    SimpleAgentState,
    make_simple_agent_server,
)
from ldp.agent.interactive_agent import InteractiveAgent
from ldp.alg import to_network
from ldp.graph import LLMCallOp, Memory, OpResult, eval_mode
from ldp.graph.gradient_estimators import llm_straight_through_estimator as llm_ste
from ldp.graph.gradient_estimators import straight_through_estimator as ste
from ldp.graph.modules import ReActModule, ToolDescriptionMethods
from ldp.llms import LLMModel

from . import CILLMModelNames
from .conftest import IN_GITHUB_ACTIONS

HERE = Path(__file__).parent


def intuitive_arg(x: str) -> float:  # type: ignore[empty-body]
    """Cast the input argument x to a float."""


class StubState(BaseModel):
    """Stub model docstring."""

    defaulted_int: int = Field(default=1, description="A description of the int.")
    required_str: str = Field(description="A description of the str.")


class StubEnum(IntEnum):
    """Stub enum docstring."""

    STUB1 = auto()
    STUB2 = auto()


def many_edge_cases(
    x: int,
    y: None,
    union: int | None,
    pydantic_model: StubState,
    basic_dict: dict[str, int],
    complex_dict: dict[str, tuple[str, int]],
    enum: StubEnum,
    defaulted_str: str = "default",
    defaulted_float: float = 1.0,
) -> None:
    """
    Check using docstrings as partial f-string templates like so: {summary_format}.

    Args:
        x: Yes, I end with a colon :
        y: I am null.
            And despite that there is a multiline argument description.
        union: I am a union and the current year is {current_year}.
        pydantic_model: I am a Pydantic model.
        basic_dict: I am a dictionary with primitive values.
        complex_dict: I am a dictionary with complex values.
        enum: I am an enum.
        defaulted_str: I have a string default value.
        defaulted_float: I have a float default value.
    """


class TestAgentState:
    @pytest.mark.vcr
    @pytest.mark.parametrize("agent", [SimpleAgent(), MemoryAgent(), ReActAgent()])
    @pytest.mark.asyncio
    async def test_no_state_mutation(self, dummy_env: DummyEnv, agent: Agent) -> None:
        obs, tools = await dummy_env.reset()

        agent_state = agent_state_0 = await agent.init_state(tools=tools)
        agent_state_0_json = agent_state_0.model_dump_json()
        for _ in range(3):  # Give a few steps to finish, the assertion needs >0 steps
            action, agent_state, _ = await agent.get_asv(agent_state, obs)
            obs, reward, done, truncated = await dummy_env.step(action.value)
            if done:
                break

        assert (
            agent_state_0_json == agent_state_0.model_dump_json()
        ), "Agent state should not be mutated between calls to get_asv"

    def test_serialization_deserializaton(self) -> None:
        orig_state = SimpleAgentState(
            messages=[Message(content="stub"), ToolRequestMessage(content="stub2")]
        )
        copied_state = SimpleAgentState(**orig_state.model_dump())
        assert orig_state.messages == copied_state.messages
        assert isinstance(copied_state.messages[1], ToolRequestMessage)


class TestSimpleAgent:
    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, CILLMModelNames.OPENAI.value]
    )
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_dummyenv(self, dummy_env: DummyEnv, model_name: str) -> None:
        obs, tools = await dummy_env.reset()

        agent = SimpleAgent(llm_model={"model": model_name, "temperature": 0.1})
        agent_state = await agent.init_state(tools=tools)
        action, agent_state, _ = await agent.get_asv(agent_state, obs)
        obs, reward, done, truncated = await dummy_env.step(action.value)
        assert (
            reward > 0
        ), "Reward should be positive, indicating agent called print_story tool"
        assert done

        # Check serialization after get_asv runs to ensure private
        # Ops aren't included
        assert agent.model_dump() == {
            "llm_model": {"model": model_name, "temperature": 0.1},
            "sys_prompt": None,
        }

        # Check we can get the LLM results to sum cost and count tokens
        assert action.call_id is not None, "Compute graph not attached to action."
        for op_r in action.traverse():
            if issubclass(op_r.op_class, LLMCallOp):
                # will raise if cannot retrieve result
                op_r._get_from_ctx("result")
                break
        else:
            raise RuntimeError("Could not find LLMCallOp in compute graph")

    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, CILLMModelNames.OPENAI.value]
    )
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_agent_grad(self, dummy_env: DummyEnv, model_name: str) -> None:
        obs, tools = await dummy_env.reset()

        agent = SimpleAgent(llm_model={"model": model_name, "temperature": 0.1})
        agent_state = await agent.init_state(tools=tools)
        action, agent_state, _ = await agent.get_asv(agent_state, obs)
        assert action.call_id is not None
        obs, reward, done, _ = await dummy_env.step(action.value)
        assert (
            reward > 0
        ), "Reward should be positive, indicating agent called print_story tool"
        assert done
        assert (
            action.call_id is not None
        ), "action is not associated with a forward pass call_id"

        # NOTE: we would not normally pass reward as a gradient, but this is a way
        # to check that gradients are flowing
        action.compute_grads(
            reward,
            backward_fns={"_config_op": ste, "_llm_call_op": llm_ste},
        )
        _, g = action.ctx.get_input_grads(action.call_id)
        assert isinstance(
            g["config"], dict
        ), "compute_grads() didn't descend into config dict"
        assert all(g["config"].values()), "Gradient should be non-zero"

        graph = to_network(action)
        with (
            tempfile.NamedTemporaryFile(mode="w", suffix=".png") as f,
            contextlib.suppress(
                FileNotFoundError  # Allow tests to run without graphviz on OS
            ),
        ):
            nx.drawing.nx_pydot.to_pydot(graph).write_png(f.name)

            if not IN_GITHUB_ACTIONS:
                output_dir = HERE / "test_outputs"
                output_dir.mkdir(exist_ok=True)
                shutil.copy(
                    f.name,
                    output_dir / f"TestSimpleAgent.test_agent_grad.{model_name}.png",
                )


class TestMemoryAgent:
    # # On 5/14/2024, claude 3 opus would not follow its past memories
    @pytest.mark.parametrize("model_name", [CILLMModelNames.OPENAI.value])
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_dummyenv(self, dummy_env: DummyEnv, model_name: str) -> None:
        obs, tools = await dummy_env.reset()

        agent = MemoryAgent(llm_model={"model": model_name, "temperature": 0.1})
        agent_state = await agent.init_state(tools=tools)

        # access memory and add one to it
        action = ToolRequestMessage(
            content="Stories that start with 'Once there was' are always interesting.",
            tool_calls=[
                ToolCall.from_name(
                    tools[0].info.name, story="Once there were was nothing."
                )
            ],
        )
        memory = agent._memory_op.memory_model
        await memory.add_memory(
            Memory(
                query="Write a 5 word story and call print",
                output=str(action),
                value=1000.0,
            )
        )

        new_action, agent_state, _ = await agent.get_asv(agent_state, obs)
        assert "Once there was" in str(new_action)
        obs, reward, done, truncated = await dummy_env.step(new_action.value)
        assert (
            reward > 0
        ), "Reward should be positive, indicating agent called print_story tool"
        assert done

        # Check we can get the LLM results to sum cost and count tokens
        assert new_action.call_id is not None, "Compute graph not attached to action."
        for op_r in new_action.traverse():
            if issubclass(op_r.op_class, LLMCallOp):
                # will raise if cannot retrieve result
                op_r._get_from_ctx("result")
                break
        else:
            raise RuntimeError("Could not find LLMCallOp in compute graph")

    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_agent_grad(self, dummy_env: DummyEnv) -> None:
        obs, tools = await dummy_env.reset()

        agent = MemoryAgent()
        agent_state = await agent.init_state(tools=tools)
        action, agent_state, _ = await agent.get_asv(agent_state, obs)
        assert action.call_id is not None
        obs, reward, done, truncated = await dummy_env.step(action.value)
        assert (
            reward > 0
        ), "Reward should be positive, indicating agent called print_story tool"
        assert done
        assert (
            action.call_id is not None
        ), "action is not associated with a forward pass call_id"

        # NOTE: we would not normally pass reward as a gradient, but this is a way
        # to check that gradients are flowing
        ste_ = partial(ste, descend=False)
        action.compute_grads(
            reward,
            backward_fns={
                "_prompt_op": ste_,
                "_package_op": ste_,
                "_format_memory_op": ste_,
                "_memory_op": ste_,
                "_llm_call_op": llm_ste,
            },
        )
        _, g = action.ctx.get_input_grads(action.call_id)
        assert isinstance(
            g["config"], dict
        ), "compute_grads() didn't descend into config dict"
        assert all(g["config"].values()), "Action gradient should be non-zero"

        memory_op = agent._memory_op
        mem_call_ids = list(memory_op.get_call_ids({action.call_id.run_id}))
        assert len(mem_call_ids) == 1, "MemoryOp should have been called exactly once"
        _, g = memory_op.get_input_grads(mem_call_ids[0])
        assert any(g.values()), "Memory gradient should be non-zero"


class TestReActAgent:
    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, "gpt-4-turbo"]
    )
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_react_dummyenv(self, dummy_env: DummyEnv, model_name: str) -> None:
        obs, tools = await dummy_env.reset()
        agent = ReActAgent(llm_model={"model": model_name, "temperature": 0.1})
        agent_state = await agent.init_state(tools=tools)
        action, agent_state, _ = await agent.get_asv(agent_state, obs)
        obs, reward, done, truncated = await dummy_env.step(action.value)
        assert (
            reward > 0
        ), "Reward should be positive, indicating agent called print_story tool"
        assert done

        # Check we can get the LLM results to sum cost and count tokens
        assert action.call_id is not None, "Compute graph not attached to action."
        for op_r in action.traverse():
            if issubclass(op_r.op_class, LLMCallOp):
                # will raise if cannot retrieve result
                op_r._get_from_ctx("result")
                break
        else:
            raise RuntimeError("Could not find LLMCallOp in compute graph")

    @pytest.mark.asyncio
    async def test_multi_step(self, dummy_env: DummyEnv) -> None:
        obs, tools = await dummy_env.reset()
        obs = dummy_env.state.messages = [
            Message(
                content=(
                    "Cast '5.5' to a float, then to an integer,"
                    " and finally use it to write a story of that many words."
                )
            )
        ]
        agent = ReActAgent()
        agent_state = await agent.init_state(tools=tools)
        for i in range(4):  # noqa: B007
            action, agent_state, _ = await agent.get_asv(agent_state, obs)
            for m in agent_state.messages:
                assert m.content
                assert (
                    "Observation: Observation" not in m.content
                ), "Prepended duplicate observations"
            obs, _, done, _ = await dummy_env.step(action.value)
            if done:
                break
        if i < 2 or not done:
            raise AssertionError(
                "Environment should have finished, with at least 2 environment steps."
            )

    def test_agent_op_naming(self) -> None:
        agent = ReActAgent()
        for op_name in (
            "prompt_op",
            "package_msg_op",
            "tool_select_module.config_op",
            "tool_select_module.llm_call_op",
            "tool_select_module.parse_msg_op",
        ):
            obj, expected = agent._react_module, f"_react_module.{op_name}"
            if "." in op_name:
                op, op_name = op_name.split(".", maxsplit=1)
                obj = getattr(obj, op)
            assert getattr(obj, op_name).name == expected

    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, "gpt-4-turbo"]
    )
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_agent_grad(self, dummy_env: DummyEnv, model_name: str) -> None:
        obs, tools = await dummy_env.reset()

        agent = ReActAgent(llm_model={"model": model_name, "temperature": 0.1})
        agent_state = await agent.init_state(tools=tools)
        action, agent_state, _ = await agent.get_asv(agent_state, obs)
        assert action.call_id is not None
        obs, reward, done, truncated = await dummy_env.step(action.value)
        assert (
            reward > 0
        ), "Reward should be positive, indicating agent called print_story tool"
        assert done
        assert (
            action.call_id is not None
        ), "action is not associated with a forward pass call_id"

        # NOTE: we would not normally pass reward as a gradient, but this is a way
        # to check that gradients are flowing
        ste_ = partial(ste, descend=False)
        action.compute_grads(
            reward,
            # Give everything a straight-through gradient approximation
            # so we can confirm gradient flow
            backward_fns={
                "_react_module.package_msg_op": ste_,
                "_react_module.prompt_op": ste_,
                "_react_module.tool_select_module.llm_call_op": llm_ste,
                "_react_module.tool_select_module.parse_msg_op": ste_,
            },
        )
        _, g = action.ctx.get_input_grads(action.call_id)
        assert all(g.values()), "Gradient should be non-zero"

        # make sure it propagated far enough
        prompt_op = agent._react_module.prompt_op
        _, g = prompt_op.get_input_grads(
            next(iter(prompt_op.get_call_ids({action.call_id.run_id})))
        )
        assert all(g.values()), "PromptOp gradients should be positive"

        graph = to_network(action, max_label_height=4, max_label_width=50)
        with (
            tempfile.NamedTemporaryFile(mode="w", suffix=".png") as f,
            contextlib.suppress(
                FileNotFoundError  # Allow tests to run without graphviz on OS
            ),
        ):
            nx.drawing.nx_pydot.to_pydot(graph).write_png(f.name)

            if not IN_GITHUB_ACTIONS:
                output_dir = HERE / "test_outputs"
                output_dir.mkdir(exist_ok=True)
                shutil.copy(
                    f.name,
                    output_dir / f"TestReActAgent.test_agent_grad.{model_name}.png",
                )

    @pytest.mark.parametrize(
        ("description_method", "expected"),
        [
            (ToolDescriptionMethods.STR, NotImplementedError),
            (
                ToolDescriptionMethods.JSON,
                (
                    "Answer the following questions as best you can. You have access to"
                    " the following tools:\n\nTools are specified with a JSON"
                    ' schema.\n{"name":"many_edge_cases","description":"Check using'
                    " docstrings as partial f-string templates like so:"
                    ' {summary_format}.","parameters":{"type":"object","properties":{"x":{"description":"Yes,'
                    " I end with a colon"
                    ' :","title":"X","type":"integer"},"y":{"description":"I am null.'
                    " And despite that there is a multiline argument"
                    ' description.","title":"Y","type":"null"},"union":{"anyOf":[{"type":"integer"},{"type":"null"}],"description":"I'
                    " am a union and the current year is"
                    ' {current_year}.","title":"Union"},"pydantic_model":{"$ref":"#/$defs/StubState","description":"I'
                    " am a Pydantic"
                    ' model."},"basic_dict":{"additionalProperties":{"type":"integer"},"description":"I'
                    ' am a dictionary with primitive values.","title":"Basic'
                    ' Dict","type":"object"},"complex_dict":{"additionalProperties":{"maxItems":2,"minItems":2,"prefixItems":[{"type":"string"},{"type":"integer"}],"type":"array"},"description":"I'
                    ' am a dictionary with complex values.","title":"Complex'
                    ' Dict","type":"object"},"enum":{"$ref":"#/$defs/StubEnum","description":"I'
                    " am an"
                    ' enum."},"defaulted_str":{"default":"default","description":"I'
                    ' have a string default value.","title":"Defaulted'
                    ' Str","type":"string"},"defaulted_float":{"default":1.0,"description":"I'
                    ' have a float default value.","title":"Defaulted'
                    ' Float","type":"number"}},"required":["x","y","union","pydantic_model","basic_dict","complex_dict","enum"],"$defs":{"StubEnum":{"description":"Stub'
                    " enum"
                    ' docstring.","enum":[1,2],"title":"StubEnum","type":"integer"},"StubState":{"description":"Stub'
                    " model"
                    ' docstring.","properties":{"defaulted_int":{"default":1,"description":"A'
                    ' description of the int.","title":"Defaulted'
                    ' Int","type":"integer"},"required_str":{"description":"A'
                    ' description of the str.","title":"Required'
                    ' Str","type":"string"}},"required":["required_str"],"title":"StubState","type":"object"}}}}\n{"name":"intuitive_arg","description":"Cast'
                    " the input argument x to a"
                    ' float.","parameters":{"type":"object","properties":{"x":{"title":"X","type":"string"}},"required":["x"]}}\n\nUse'
                    " the following format:\n\nThought: you should always think about"
                    " what to do\nAction: the action to take, should be one of"
                    " [many_edge_cases, intuitive_arg]\nAction Input: comma separated"
                    " list of inputs to action as python tuple\nObservation: the result"
                    " of the action\n... (this Thought/Action/Action Input/Observation"
                    " can repeat N times)\n\nExample:\n\nThought: I need to use the"
                    ' get_weather tool\nAction: get_weather\nAction Input: "New York",'
                    " 7\nObservation: The 7 day forecast for New York is [...]"
                ),
            ),
            (
                ToolDescriptionMethods.XML,
                (
                    "Answer the following questions as best you can. You have access to"
                    " the following tools:\n\nTools are specified with an XML"
                    " schema.\n<function_info><name>many_edge_cases</name><description>Check"
                    " using docstrings as partial f-string templates like so:"
                    " {summary_format}.</description><parameters><type>object</type><properties><x><description>Yes,"
                    " I end with a colon"
                    " :</description><title>X</title><type>integer</type></x><y><description>I"
                    " am null. And despite that there is a multiline argument"
                    " description.</description><title>Y</title><type>null</type></y><union><anyOf><item><type>integer</type></item><item><type>null</type></item></anyOf><description>I"
                    " am a union and the current year is"
                    " {current_year}.</description><title>Union</title></union><pydantic_model><key"
                    ' name="$ref">#/$defs/StubState</key><description>I am a Pydantic'
                    " model.</description></pydantic_model><basic_dict><additionalProperties><type>integer</type></additionalProperties><description>I"
                    " am a dictionary with primitive values.</description><title>Basic"
                    " Dict</title><type>object</type></basic_dict><complex_dict><additionalProperties><maxItems>2</maxItems><minItems>2</minItems><prefixItems><item><type>string</type></item><item><type>integer</type></item></prefixItems><type>array</type></additionalProperties><description>I"
                    " am a dictionary with complex values.</description><title>Complex"
                    " Dict</title><type>object</type></complex_dict><enum><key"
                    ' name="$ref">#/$defs/StubEnum</key><description>I am an'
                    " enum.</description></enum><defaulted_str><default>default</default><description>I"
                    " have a string default value.</description><title>Defaulted"
                    " Str</title><type>string</type></defaulted_str><defaulted_float><default>1.0</default><description>I"
                    " have a float default value.</description><title>Defaulted"
                    " Float</title><type>number</type></defaulted_float></properties><required><item>x</item><item>y</item><item>union</item><item>pydantic_model</item><item>basic_dict</item><item>complex_dict</item><item>enum</item></required><key"
                    ' name="$defs"><StubEnum><description>Stub enum'
                    " docstring.</description><enum><item>1</item><item>2</item></enum><title>StubEnum</title><type>integer</type></StubEnum><StubState><description>Stub"
                    " model"
                    " docstring.</description><properties><defaulted_int><default>1</default><description>A"
                    " description of the int.</description><title>Defaulted"
                    " Int</title><type>integer</type></defaulted_int><required_str><description>A"
                    " description of the str.</description><title>Required"
                    " Str</title><type>string</type></required_str></properties><required><item>required_str</item></required><title>StubState</title><type>object</type></StubState></key></parameters></function_info>\n<function_info><name>intuitive_arg</name><description>Cast"
                    " the input argument x to a"
                    " float.</description><parameters><type>object</type><properties><x><title>X</title><type>string</type></x></properties><required><item>x</item></required></parameters></function_info>\n\nUse"
                    " the following format:\n\nThought: you should always think about"
                    " what to do\nAction: the action to take, should be one of"
                    " [many_edge_cases, intuitive_arg]\nAction Input: comma separated"
                    " list of inputs to action as python tuple\nObservation: the result"
                    " of the action\n... (this Thought/Action/Action Input/Observation"
                    " can repeat N times)\n\nExample:\n\nThought: I need to use the"
                    ' get_weather tool\nAction: get_weather\nAction Input: "New York",'
                    " 7\nObservation: The 7 day forecast for New York is [...]"
                ),
            ),
        ],
    )
    @pytest.mark.asyncio
    async def test_complex_system_prompt(
        self,
        description_method: ToolDescriptionMethods,
        expected: str | type[Exception],
    ) -> None:
        tools = [
            Tool.from_function(many_edge_cases),
            Tool.from_function(intuitive_arg, allow_empty_param_descriptions=True),
        ]
        user_msg = Message(content="Cast the string '5.6' to a float.")
        with (
            patch.object(LLMModel, "achat") as mock_achat,
            patch.object(ReActModule, "parse_message"),
        ):
            agent = ReActAgent(tool_description_method=description_method)
            agent_state = await agent.init_state(tools=tools)
            if not isinstance(expected, str):
                with pytest.raises(expected):
                    await agent.get_asv(agent_state, obs=[user_msg])
                return
            await agent.get_asv(agent_state, obs=[user_msg])
        mock_achat.assert_awaited_once()
        assert mock_achat.await_args
        assert mock_achat.await_args[0][0] == [
            Message(role="system", content=expected),
            user_msg,
        ]


class TestHTTPAgentClient:
    @patch.dict("os.environ", {"AUTH_TOKEN": "stub"})
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_lifecycle(self, dummy_env: DummyEnv) -> None:
        obs, tools = await dummy_env.reset()
        # Let's turn the prompt to require multiple steps
        obs = dummy_env.state.messages = [
            Message(
                content=(
                    "Cast '5.5' to a float, then to an integer,"
                    " and finally use it to write a story of that many words."
                )
            )
        ]

        remote_agent = SimpleAgent()
        base_server_url = "http://testserver"
        agent_client = HTTPAgentClient[SimpleAgentState](
            agent_state_type=SimpleAgentState,
            server_url=base_server_url,
            request_headers={"Authorization": "Bearer stub"},
        )
        async with AsyncClient(
            transport=ASGITransport(app=make_simple_agent_server(agent=remote_agent)),
            base_url=base_server_url,
        ) as async_client:
            # NOTE: just directly hit the server, since info is not an Agent method
            response = await async_client.get(
                f"{base_server_url}/info", headers={"Authorization": "Bearer stub"}
            )
            response.raise_for_status()
            assert response.json()["agent_type"] == "SimpleAgent"

            with patch("httpx.AsyncClient.post", async_client.post):
                agent_state_0 = await agent_client.init_state(tools=tools)
            assert [t.info for t in agent_state_0.tools] == [t.info for t in tools]

            # NOTE: also check we can repeatedly call the get_asv
            with patch("httpx.AsyncClient.post", async_client.post), eval_mode():
                await agent_client.get_asv(agent_state_0, obs)
                await agent_client.get_asv(agent_state_0, obs)
            with patch("httpx.AsyncClient.post", async_client.post), eval_mode():
                action, agent_state_1, vhat = await agent_client.get_asv(
                    agent_state_0, obs
                )

            assert isinstance(action, OpResult)
            assert isinstance(action.value, ToolRequestMessage)
            assert isinstance(agent_state_1, SimpleAgentState)
            assert len(agent_state_1.messages) == 2
            assert isinstance(agent_state_1.messages[0], Message)
            assert isinstance(agent_state_1.messages[1], ToolRequestMessage)
            assert isinstance(vhat, float)

            # This makes an obs with ToolResponseMessage inside
            obs, reward, done, _ = await dummy_env.step(action.value)
            assert not done

            with patch("httpx.AsyncClient.post", async_client.post), eval_mode():
                # Check we can make a second sequential Agent decision without crashing
                await agent_client.get_asv(agent_state_1, obs)


@pytest.mark.parametrize("agent_cls", [SimpleAgent, MemoryAgent, ReActAgent])
def test_agent_config(agent_cls: type[Agent]):
    config = AgentConfig(agent_type=agent_cls.__name__)
    assert isinstance(hash(config), int), "AgentConfig should be hashable"
    agent = config.construct_agent()
    assert isinstance(agent, agent_cls)


@pytest.mark.asyncio
async def test_interactive(dummy_env: DummyEnv, mocker):
    agent = InteractiveAgent()

    obs, tools = await dummy_env.reset()
    agent_state = await agent.init_state(tools=tools)
    done = False

    mock_input = mocker.patch(
        "builtins.input", side_effect=["print_story", "A cat wore a hat."]
    )

    action, agent_state, _ = await agent.get_asv(agent_state, obs)
    next_obs, _, done, _ = await dummy_env.step(action.value)

    assert mock_input.call_count >= 2



from __future__ import annotations

import asyncio
import logging
from collections import UserDict
from enum import StrEnum, auto
from typing import Any, Self, cast

from aviary.message import Message
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    JsonValue,
    RootModel,
    ValidationError,
    model_validator,
)

from ldp.agent import ReActAgent
from ldp.alg.optimizer.opt import Optimizer
from ldp.data_structures import Trajectory
from ldp.graph import LLMCallOp, OpResult, PromptOp
from ldp.llms import LLMModel, LLMResult

logger = logging.getLogger(__name__)


class APEScoreFn(StrEnum):
    # Use the reward as the APE score (as proposed in the paper).
    # Goal is to maximize this score.
    REWARD = auto()
    # Use the gradient of the output of the PromptOp as the APE score.
    # Goal is to push this to zero.
    GRADIENT = auto()


class _FormatDict(UserDict):
    """Custom dictionary that stores missing items."""

    def __init__(self) -> None:
        super().__init__()
        self.key_set: set[str] = set()

    def __missing__(self, key: str) -> str:
        self.key_set.add(key)
        return key


def get_formatted_variables(s: str) -> set[str]:
    """Returns the set of variables implied by the format string."""
    format_dict = _FormatDict()
    s.format_map(format_dict)
    return format_dict.key_set


class OutputPrompt(BaseModel):
    prompt: str = Field(description="Prompt for language model")


class Example(BaseModel):
    input: JsonValue
    output: JsonValue
    score: float


ExampleList = RootModel[list[Example]]


class APEOpt(BaseModel, Optimizer):
    """
    Basic optimizer that acts as an Automatic Prompt Engineer (APE).

    Paper: https://openreview.net/pdf?id=92gvk82DE-

    Details:
    - This implements the "forward mode generation" strategy.
    - The score function used is the gradient (float) at the output of the
      PromptOp being optimized. A zero gradient means the prompt was "good",
      and a non-zero gradient means we can learn from the prompt.
    - Possible improvements include:
        - Extending the score function to the LLM result's logprobs
        - Iterating with Monte Carlo Search
        - Use of memory for further example augmentation
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    ### Configuration
    max_examples: int | None = Field(
        default=50,  # Comes from APE paper section 5.2
        description=(
            "Max number of examples to include in the below query_prompt, or None for"
            " no limit. The paper mentions that more examples produce better prompt"
            " proposals."
        ),
    )
    system_prompt: str = (
        "We are optimizing prompts for a language model."
        " The model sees a prompt, an input, and then generates an output."
    )
    query_prompt: str = (
        "Here are correct example outputs that the language model and prompt should produce:"
        "\n{good_examples}"
        '\n\nThe current prompt is: "{prompt}"'
        "\n\nWhich resulted in the following incorrect input, output, and {score}:"
        "\n{examples}"
        "\n\nRevise the current prompt to improve the outputs."
        " Your proposed prompt should be concise, correct, and specify the desired output format."
    )
    llm: LLMModel = Field(
        default_factory=LLMModel,
        description=(
            "LLM used to update the prompt inside the PromptOp. The paper mentions that"
            " larger models produce better prompt proposals."
        ),
    )

    prompt_op: PromptOp = Field(description="PromptOp to be optimized.")
    llm_call_op: LLMCallOp = Field(description="LLMCallOp to be optimized.")

    score_fn: APEScoreFn = APEScoreFn.REWARD
    good_reward_threshold: float | None = Field(
        default=None,
        description="If using reward as the score_fn, then a good example is defined by "
        "reward>=good_reward_threshold.",
    )
    reward_discount: float = 1.0

    ### State
    examples: list[Example] = Field(default_factory=list)
    good_examples: list[Example] = Field(default_factory=list)
    steps: int = 0
    trace: list[str] = Field(
        default_factory=list, description="History of prompts used."
    )

    @model_validator(mode="after")
    def validate_score_fn(self):
        if self.score_fn == APEScoreFn.REWARD:
            if self.good_reward_threshold is None:
                raise ValueError(
                    "good_reward_threshold must be set if using reward as the score function"
                )
            self._score_str = "rewards"
        elif self.score_fn == APEScoreFn.GRADIENT:
            # The gradient into the prompt op is the (signed) backpropagated error, and "gradient" would
            # be confusing to the model in the prompt.
            self._score_str = "errors"
        else:
            raise ValueError(f"Invalid score function {self.score_fn}")

        return self

    def model_post_init(self, __context: Any) -> None:
        if self.prompt_op.prompt not in self.trace:
            self.trace.append(self.prompt_op.prompt)

        # Make sure updates are not run concurrently
        self._update_lock = asyncio.Lock()

    @classmethod
    def from_agent(cls, agent: ReActAgent, **kwargs) -> Self:
        return cls(
            llm_call_op=agent._react_module.tool_select_module.llm_call_op,
            prompt_op=agent._react_module.prompt_op,
            **kwargs,
        )

    def aggregate_trajectory(self, trajectory: Trajectory) -> None:
        if trajectory.failed:
            return

        if self.score_fn == APEScoreFn.REWARD:
            d_returns = trajectory.compute_discounted_returns(self.reward_discount)

        for i_step, step in enumerate(trajectory.steps):
            action = cast(OpResult, step.action)

            if self.score_fn == APEScoreFn.GRADIENT:
                prompt_op_result, *extra = action.get_upstream_results(self.prompt_op)
                assert not extra, "APE only supports one prompt call per run"

            for op_result in action.get_upstream_results(self.llm_call_op):
                result = cast(
                    LLMResult | None,
                    self.llm_call_op.ctx.get(op_result.call_id, "result"),
                )
                if result is None or not result.messages or not result.prompt:
                    continue
                # (x: first prompt's user message's content, y: AI response's content)
                x = next(
                    cast(str, m.content) for m in result.prompt if m.role == "user"
                )
                y = cast(str, result.messages[0].content)

                if self.score_fn == APEScoreFn.GRADIENT:
                    score = self.prompt_op.ctx.get(
                        prompt_op_result.call_id, "grad_output", default=None
                    )
                    if score is None:
                        # backprop did not reach this op call - move on
                        continue
                    is_good = score == 0
                else:
                    score = d_returns[i_step]  # pylint: disable=possibly-used-before-assignment
                    is_good = score >= cast(float, self.good_reward_threshold)

                example = Example(input=x, output=y, score=score)
                (self.good_examples if is_good else self.examples).append(example)

    async def update(self) -> None:
        async with self._update_lock:
            if not self.examples:
                raise ValueError("No examples to update the prompt with.")

            new_p = await self._get_updated_prompt(
                self.examples, self.good_examples, self.prompt_op.prompt
            )
            # Check any template vars remain, and if some were added or
            # lost, discard this new prompt
            if new_p != self.prompt_op.prompt and get_formatted_variables(
                new_p
            ) != get_formatted_variables(self.prompt_op.prompt):
                logger.warning(
                    "Update broke prompt templating."
                    f"\n\nNew prompt:\n{new_p}"
                    f"\n\nPrior prompt:\n{self.prompt_op.prompt}"
                )
            else:
                if new_p == self.prompt_op.prompt:
                    logger.warning("Update did not change the prompt.")
                self.examples.clear()
                self.prompt_op.prompt = new_p
                self.trace.append(new_p)
            self.steps += 1

    def _prepare_examples(self, examples: list[Example]) -> str:
        if not examples:
            return ""
        if self.max_examples and len(examples) > self.max_examples:
            if self.score_fn == APEScoreFn.GRADIENT:
                # Return examples closest to decision boundary,
                # aka ones with the lowest L1-normalized error
                # NOTE: this pairs with our setting of Example.score = PromptOp's output
                # gradient inside the update method, so examples with error values closer to
                # 0 are defined to be higher quality
                examples = sorted(examples, key=lambda e: abs(e.score))[
                    : self.max_examples
                ]
            else:
                # In reward mode, we want to show the examples with the highest reward, per the paper
                # TODO: consider whether uniform sampling is better
                examples = sorted(examples, key=lambda e: -e.score)[: self.max_examples]
        return ExampleList.model_validate(examples).model_dump_json()

    async def _get_updated_prompt(
        self, examples: list[Example], good_examples: list[Example], prompt: str
    ) -> str:
        messages = [
            Message(
                role="system",
                content=self.system_prompt,
            ),
            Message(
                role="user",
                content=self.query_prompt.format(
                    examples=self._prepare_examples(examples),
                    good_examples=self._prepare_examples(good_examples),
                    prompt=prompt,
                    score=self._score_str,
                ),
            ),
            Message(
                role="assistant",
                content=OutputPrompt(prompt=prompt).model_dump_json(indent=2),
            ),
            Message(
                content=(
                    "You responded without changing the prompt. Don't forget to revise"
                    " the prompt."
                )
            ),
        ]
        result = await self.llm.call(messages, output_type=OutputPrompt)
        message_content = cast(str, cast(list[Message], result.messages)[-1].content)
        try:
            return OutputPrompt.model_validate_json(message_content).prompt
        except ValidationError:
            return prompt



import random
from collections import UserList


class CircularReplayBuffer(UserList[dict]):
    def resize(self, size: int):
        if len(self) > size:
            self.data = self.data[-size:]

    def batched_iter(
        self, batch_size: int, shuffle: bool = True, infinite: bool = False
    ):
        while True:
            indices = list(range(len(self)))
            if shuffle:
                random.shuffle(indices)

            for i in range(0, len(self), batch_size):
                keys = self.data[0].keys()

                batch: dict[str, list] = {k: [] for k in keys}
                for j in indices[i : i + batch_size]:
                    if self.data[j].keys() != keys:
                        raise RuntimeError(
                            "Found buffer element with inconsistent keys"
                        )

                    for k in keys:
                        batch[k].append(self.data[j][k])

                yield batch

            if not infinite:
                break
